#include "textmode.h"
#include "version.h"
#include "../lib/stdlib.h"
#include "../lib/port.h"

/**
 * A structure for storing video mode information.
 */
static struct ktextmode_descr {
    uQWord mem_address;     // Memory offset address
    uByte lines;            // Max lines
    uByte columns;          // Max columns
    uByte cur_line;         // Current line
    uByte cur_pos;          // Current position
    uByte cur_fgCol;        // Foreground
    uByte cur_bgCol;        // Background
} ktextmode_cur = {
    TEXTMODE_MEMBASE,
    25,
    80,
    0,
    0,
    TEXTMODE_LIGHTGREY,
    TEXTMODE_BLACK
};

/**
 * Little helper inline function which calculates the higher attribute byte
 * for text output.
 *
 * @return A uWord containing the attribute byte in the upper region.
 */
static inline uWord ktextmode_calcAttr() {
    return ((ktextmode_cur.cur_fgCol & 0xF) |
            (ktextmode_cur.cur_bgCol << 4)) << 8;
}

/**
 * Updates the cursor position in text mode.
 */
static void kupdate_cursor() {
   uWord loc = ktextmode_cur.cur_line * ktextmode_cur.columns +
               ktextmode_cur.cur_pos;
   kport_outb(0x3D4, 14);
   kport_outb(0x3D5, loc >> 8);
   kport_outb(0x3D4, 15);
   kport_outb(0x3D5, loc);
}

/**
 * Scrolls the text up by one line.
 */
static void kscroll() {
    int i;
    int max = (ktextmode_cur.lines - 1) * ktextmode_cur.columns;
    uWord* mem = (uWord*) ktextmode_cur.mem_address;

    for(i = 0; i < max; ++i) {
        mem[i] = mem[i + ktextmode_cur.columns];
    }

    max = ktextmode_cur.lines * ktextmode_cur.columns;

    for(; i < max; ++i) {
        mem[i] &= 0xFF00;
    }
}

/**
 * Clears the screen in text mode.
 */
void kclear_screen() {
    uWord max = ktextmode_cur.columns * ktextmode_cur.lines;
    uWord* dest = (uWord*) ktextmode_cur.mem_address;
    uWord i;
    uWord attr = ktextmode_calcAttr();
    ktextmode_cur.cur_line = 0;
    ktextmode_cur.cur_pos = 0;

    for(i = 0; i < max; ++i) {
        dest[i] = attr;
    }
}

void kset_textcol(uByte fgCol, uByte bgCol) {
    ktextmode_cur.cur_fgCol = fgCol;
    ktextmode_cur.cur_bgCol = bgCol;
}

/**
 * Prints a zero-terminated string to the screen at the specified position.
 *
 * @param txt The message to print
 */
void kprint(const char* txt) {
    int idx = 0;
    uWord* dest;
    uWord attr = ktextmode_calcAttr();

    while(txt[idx]) {
        switch(txt[idx]) {
        case '\n':
            ktextmode_cur.cur_line++;

        case '\r':
            ktextmode_cur.cur_pos = 0;
            break;

        case '\t':
            ktextmode_cur.cur_pos = (ktextmode_cur.cur_pos + 8) & ~7;
            break;

        default:
            dest = (uWord*) (ktextmode_cur.mem_address +
                    (ktextmode_cur.cur_line * ktextmode_cur.columns +
                     ktextmode_cur.cur_pos) * 2);
            *dest = txt[idx] | attr;
            ++ktextmode_cur.cur_pos;

            if(ktextmode_cur.cur_pos == ktextmode_cur.columns) {
                ktextmode_cur.cur_pos = 0;
                ++ktextmode_cur.cur_line;
            }
        }

        if(ktextmode_cur.cur_line == ktextmode_cur.lines) {
            kscroll();
            ktextmode_cur.cur_line--;
        }

        ++idx;
    }

    kupdate_cursor();
}

/**
 * Repeatedly prints the character up to the end of the current line.
 *
 * @param chr The character to print repeatedly.
 */
void kprint_fill(char chr) {
    int i;
    char out[2] = {chr, 0};

    for(i = ktextmode_cur.cur_pos; i < ktextmode_cur.columns; ++i) {
        kprint((char*) &out);
    }
}

/**
 * Prints a numeric value to screen.
 *
 * @param number The number to print.
 * @param base   The output base.
 */
void kprint_int(int number, uByte base) {
    // Binary representation has a maximum of 64 characters + sign + 0
    char str[66];
    itoa(number, str, base);
    kprint(str);
}

inline void kprint_hex(int number) {
    kprint_int(number, 16);
}

inline void kprint_dec(int number) {
    kprint_int(number, 10);
}

inline void kprint_oct(int number) {
    kprint_int(number, 8);
}

inline void kprint_bin(int number) {
    kprint_int(number, 2);
}

void kprint_kernel_version() {
    kprint("v");
    kprint_dec(BELLA_VERSION_MAJOR);
    kprint(".");
    kprint_dec(BELLA_VERSION_MINOR);
}

void kprint_kernel_compiledatetime() {
    kprint(__DATE__);
    kprint(", ");
    kprint(__TIME__);
}
