#include "stdio.h"
#include "stdlib.h"
#include "../kernel/textmode.h"

#define JUSTIFY_LEFT    1
#define FORCE_SIGN      2
#define FORCE_BLANK     4
#define ZEROX_PREFIX    8
#define ZERO_PAD        16
#define ARG_PRECISION   32
#define CAPITALS        64

#define LEN_SHORT    1
#define LEN_LONG     2
#define LEN_DOUBLE   3

/**
 * Prints a formatted list of input variables to the target string.
 *
 * @param str The target string buffer.
 * @param format A string containing the (formatting) output.
 * @param args An argument list data structure from another function call.
 *
 * @return The string length.
 */
int vsprintf(char* str, const char* format, va_list args) {
    int cnt;
    int len = 0;
    int fieldWidth;
    int precision;
    uByte base;
    char specLen;
    char* endptr;
    char* argstr;
    char tmp_str[66];
    char* p_tmp_str;
    uByte flags;

    while(*format) {
        base = 0;
        flags = 0;
        precision = 0;
        specLen = 0;
        fieldWidth = 0;

        // Format tag detected
        if(*format == '%') {
            if(*++format == '%') {
                *str++ = '%';
                ++len;
                continue;
            }

            // Detect optional format flags
            while(*format) {
                switch(*format++) {
                case '-':
                    flags |= JUSTIFY_LEFT;
                    continue;

                case '+':
                    flags |= FORCE_SIGN;
                    continue;

                case ' ':
                    flags |= FORCE_BLANK;
                    continue;

                case '#':
                    flags |= ZEROX_PREFIX;
                    fieldWidth -= 2;
                    continue;

                case '0':
                    flags |= ZERO_PAD;
                    continue;
                }

                --format;
                break;
            }

            fieldWidth += strtol(format, &endptr, 10);
            format = endptr;

            // Detect precision
            if(*format == '.') {
                if(*(++format) == '*') {
                    flags |= ARG_PRECISION;
                    ++format;
                } else {
                    precision = strtol(format, &endptr, 10);
                    format = (char*) endptr;
                }
            }

            switch(*format) {
            case 'h':
            case 'l':
            case 'L':
                specLen = *format;
                ++format;
            }

            // Execute formatted string output now:
            switch(*format++) {
            case 'c':
                break;

            case 'd':
            case 'i':
                flags |= FORCE_SIGN;
                base = 10;
                break;

            case 'e':
            case 'E':
                break;

            case 'f':
                break;

            case 'g':
                break;

            case 'G':
                break;

            case 'o':
                base = 8;
                flags |= FORCE_SIGN;
                break;

            case 's':
                argstr = va_arg(args, char*);

                while(*argstr) {
                    *str++ = *argstr++;
                    ++len;
                }
                break;

            case 'u':
                base = 10;
                flags &= ~FORCE_SIGN;
                break;

            case 'X':
                flags |= CAPITALS;
            case 'x':
                base = 16;
                flags &= ~FORCE_SIGN;
                break;

            case 'p':
                break;

            case 'n':
                break;
            }

            if(base) {
                uByte nts_flags = 0;
                p_tmp_str = (char*) &tmp_str[0];

                if(flags & FORCE_SIGN) nts_flags |= NUMTOSTR_SIGNED;
                if(!(flags & CAPITALS)) nts_flags |= NUMTOSTR_LOWERCASE;
                cnt = numtostr_core(va_arg(args, int), p_tmp_str,
                                    base, nts_flags, precision);

                while(fieldWidth-- > cnt) {
                    *str++ = ' ';
                    ++len;
                }

                if(flags & ZEROX_PREFIX) {
                    *str++ = '0';

                    if(flags & CAPITALS) {
                        *str++ = 'X';
                    } else {
                        *str++ = 'x';
                    }

                    len += 2;
                }

                while(*p_tmp_str) {
                    *str++ = *p_tmp_str++;
                }

                len += cnt;
            }
        } else { // Normal character
            *str++ = *format;
            ++len;
            ++format;
        }
    }

    *str = 0;

    return len;
}

/**
 * Prints a formatted list of input variables to the target string.
 *
 * @param str The target string buffer.
 * @param format A string containing the (formatting) output.
 * @param args A variable length argument list of values to print in a formatted
 *        manner.
 *
 * @return The string length.
 */
int sprintf(char* str, const char* format, ...) {
    va_list args;
    va_start(args, format);
    int len = vsprintf(str, format, args);
    va_end(args);

    return len;
}

/**
 * A kernel function for direct formatted printing to the standard output.
 *
 * @param format The format string.
 * @param ... A variable length list of arguments to format.
 *
 * @return The amount of bytes which were printed.
 */
int kprintf(const char* format, ...) {
    // TODO: Check whether there's a more elegant way
    char buf[1024];
    va_list args;
    va_start(args, format);
    int len = vsprintf(buf, format, args);
    va_end(args);
    kprint(buf);

    return len;
}
