#include "stdlib.h"
#include "string.h"

/**
 * A helper function which converts a long integer value to its string
 * representation.
 *
 * @param value The integer value to convert.
 * @param str   The target string buffer.
 * @param base  The numeric base to convert to.
 * @param flags Possible bit flags are NUMTOSTR_SIGNED and NUMTOSTR_LOWERCASE.
 */
int numtostr_core(uQWord value, char* str, int base, uByte flags,
                  int precision) {
    char big_digits[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    char small_digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
    char* digits;
    int len = 0;
    int marker = 0;
    uQWord div = value;

    if(flags & NUMTOSTR_LOWERCASE) {
        digits = (char*) &small_digits;
    } else {
        digits = (char*) &big_digits;
    }

    if((flags & NUMTOSTR_SIGNED) && ((sQWord) value < 0)) {
        value = -value;
        *str++ = '-';
        ++len;
    }

    marker = len;

    do {
        ++str;
        ++len;
    } while(div /= base);

    while(precision > len) {
        ++str;
        ++len;
    }

    *str = 0;
    marker = len - marker;

    do {
        *--str = digits[value % base];
    } while(value /= base);

    while(marker-- > 0) {
        *--str = '0';
    }

    return len;
}

/**
 * Converts a (signed) integer value into its string representation
 * under a given numeric base (e.g. 2, 8, 10, 16).
 *
 * @param value The number to convert to a string.
 * @param str   The target string buffer.
 * @param base  The target numeric system's base.
 * @return The string length in bytes.
 */
int itoa(int value, char* str, int base) {
    return numtostr_core((uQWord) value, str, base, NUMTOSTR_SIGNED, 0);
}

/**
 *
 */
int uitoa(unsigned int value, char* str, int base) {
    return numtostr_core((uQWord) value, str, base, 0, 0);
}

/**
 * Converts a floating point value to its base 10 string representation.
 *
 * @param value     The float to convert.
 * @param str       The target string buffer.
 * @param presision The number of digits to follow after the period.
 * @return The string length in bytes.
 *
 * FIXME: Doesn't work, yet. For some reason the passed pointer address is shifted by 32 bytes towards the origin. Why?!
 */
int ftoa(float value, char* str, uByte precision) {
    sByte sign = (value < 0.0f) ? -1 : 1;
    int intVal = (int) value;
    float slide = ((float) sign * value) - (float) intVal;
    int i;

    for(i = 0; i < precision; ++i) {
        slide *= 10;
    }

    int fract = (int) (slide + 0.5 * sign);
    int len = itoa(intVal, str, 10);
    str[len++] = '.';
    len += itoa(fract, (char*) &str[len], 10);

    return len;
}

/**
 * Converts a string representation of a number of the specified base
 * to its according signed long integer value.
 *
 * @param str       The string to convert to a number.
 * @param endptr    Will be updated to the address position where
 *                  string interpretation ended.
 * @param base      The numeric base (e.g. 2, 8, 10, 16).
 * @return The converted value.
 */
long int strtol(const char* str, char** endptr, int base) {
    long int result = 0;
    long int fact = 1;
    sByte sign = 1;
    char minLetter = 0;
    char maxLetter = 0;
    char digit;
    uByte endpos = 0;
    uByte startpos = 0;

    if(base > 10) {
        minLetter = 'A';
        maxLetter = 'A' + (char) base - 10;
    }

    while(is_whitespace(*str)) {
        ++str;
    }

    if(*str == '-') {
        ++str;
        sign = -1;
        startpos++;
    }

    endpos = startpos;

    while(*str) {
        digit = chrupr(*str++);

        if(is_digit(digit) || ((digit >= minLetter) && (digit <= maxLetter))) {
            endpos++;
        } else {
            --str;
            break;
        }
    }

    *endptr = (char*) str;

    while(endpos-- > startpos) {
        if(is_digit(digit = chrupr(*--str))) {
            digit -= '0';
        } else {
            digit -= ('A' - 10);
        }

        result += fact * (int) digit;
        fact *= base;
    }

    return sign * result;
}

/**
 * Converts a string representation to its according integer value (base 10).
 *
 * @param str The string to convert.
 * @return The converted value.
 */
int atoi(const char* str) {
    char* dummy;
    return strtol(str, &dummy, 10);
}
